"""Configure the logger for the whole project.

To use it, add the following lines at the top of your file:

from biodatasets.utils import logger

log = logger.get(__name__)

The logger can be used in the file by calling :
- log.debug('My debug message')
- log.info('My info message')
- log.warning('My warning message')
- log.critical('My critical message')
- log.exception('My exception message')

The last log can be used in try / except statement and it will log the full traceback

By default the log level is set to INFO, if you want to run as DEBUG you can do the following:
LOGLEVEL=DEBUG python ....
"""
import logging
import os


def get(name: str) -> logging.Logger:
    """Create a logger for the given module name.

    Args:
        name: module name used to called the 'get' method

    Returns:
        the logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(os.environ.get("LOGLEVEL", "INFO").upper())

    handler = logging.StreamHandler()
    formatter = logging.Formatter("%(levelname)s: %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False

    return logger



from pathlib import Path
from typing import List
from typing import Optional
from typing import Union
from urllib.request import urlretrieve

from google.cloud import storage
from google.cloud.storage import Blob
from tqdm import tqdm

from biodatasets import CACHE_DIRECTORY
from biodatasets.utils import logger

log = logger.get(__name__)

DATASET_BUCKET_NAME = "deepchain-datasets-public"


def list_blobs() -> List[Blob]:
    """Lists all the blobs for a bucket."""

    storage_client = storage.Client.create_anonymous_client()
    datasets_bucket = storage_client.get_bucket(DATASET_BUCKET_NAME)
    blobs = datasets_bucket.list_blobs()

    return blobs


def get_dataset_path(name: str) -> Path:
    """Get dataset path.

    Args:
        name: name of the dataset

    Returns:
        Path to cache directory of the dataset
    """
    return CACHE_DIRECTORY / name


def pull_dataset(name: str, force: bool = False) -> None:
    """Pull data from the Google Bucket.

    Args:
        name: name of the dataset
        force: force the download of the dataset
    """

    cache_dataset_directory = get_dataset_path(name)
    storage_client = storage.Client.create_anonymous_client()
    datasets_bucket = storage_client.get_bucket(DATASET_BUCKET_NAME)

    log.info(f"Start downloading {name} dataset in {cache_dataset_directory} ...")
    file_paths = [blob.name for blob in datasets_bucket.list_blobs(prefix=f"{name}/")]

    for file_path in file_paths:
        # filter on empty dataset folder path, in case the folder has been manually created
        if file_path[-1] != "/":
            download_from_bucket(datasets_bucket, file_path, force=force)


def download_from_bucket(
    bucket: storage.Bucket,
    bucket_file_path: str,
    local_file_path: Optional[Union[Path, str]] = None,
    force: bool = False,
) -> None:
    """
    Download the file from the bucket to the local machine.

    If the local_directory is specified the files are downloaded to this directory,
    otherwise the structure of the file path is preserved.

    Args:
        bucket: bucket from which to download the file
        bucket_file_path: file path in the bucket
        local_file_path: path to which we save locally
        force: whether to force the download or not

    Raises:
        FileNotFoundError: if the file does not exist.
    """
    gs_blob = bucket.blob(bucket_file_path)

    if not gs_blob.exists():
        raise FileNotFoundError(
            f"The file {bucket_file_path} does not exist in Google Bucket '{bucket.name}'"
        )

    if local_file_path is None:
        local_file_path = CACHE_DIRECTORY / bucket_file_path
    else:
        local_file_path = _convert_file_path(local_file_path).resolve()

    should_download = force or not local_file_path.exists()

    if should_download:
        local_file_path.parent.mkdir(exist_ok=True, parents=True)
        url = gs_blob.public_url
        filename = url.split("/")[-1]
        # gs_blob.download_to_filename(local_file_path) no progress bar

        with TqdmUpTo(unit="B", unit_scale=True, unit_divisor=1024, miniters=1, desc=filename) as t:
            urlretrieve(url, filename=local_file_path, reporthook=t.update_to)

        log.info(
            f"File {bucket_file_path} downloaded from Google Bucket "
            f"'{bucket.name}' at {local_file_path}"
        )


def _convert_file_path(file_path: Union[str, Path]) -> Path:
    """Optionally convert the file path into Path.

    Args:
        file_path: file path to convert

    Return:
        the converted file path
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)

    return file_path


class TqdmUpTo(tqdm):
    """Alternative class-based version of the above.

    Provides `update_to(n)` which uses `tqdm.update(delta_n)`.
    Inspired by [twine#242](https://github.com/pypa/twine/pull/242),
    """

    def update_to(self, b: int = 1, bsize: int = 1, tsize: Optional[int] = None):
        """Update function.

        Args:
            b  : int, optional
                Number of blocks transferred so far [default: 1].
            bsize  : int, optional
                Size of each block (in tqdm units) [default: 1].
            tsize  : int, optional
                Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize



"""
This module aims to define the Dataset class and the load_dataset api function.
"""
from pathlib import Path
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import numpy as np
import pandas as pd

from biodatasets.utils import logger
from biodatasets.utils.google_bucket import get_dataset_path
from biodatasets.utils.google_bucket import list_blobs
from biodatasets.utils.google_bucket import pull_dataset

log = logger.get(__name__)


class Dataset:
    """Dataset class."""

    def __init__(self, name: str, force: bool = False):
        """Init the Dataset instance by fetching the dataset files.

        Args:
            name: name of the dataset
            force: force fetching the dataset
        """
        self.name = name
        self.path = get_dataset_path(name)
        if force or not self.path.exists():
            pull_dataset(name, force=force)

    def __repr__(self) -> str:
        """String representation of the dataset."""
        return f"Dataset(name={self.name}, available_columns={self.available_columns}, available_embeddings={self.available_embeddings})"

    @property
    def csv_path(self) -> Path:
        """Path to the CSV file."""
        return self.path / "dataset.csv"

    @property
    def description_path(self) -> Path:
        """Path to the description.md file."""
        return self.path / "description.md"

    @property
    def available_embeddings(self) -> Set[Tuple[str, str, str]]:
        """Return the available embeddings (column_name, model_name, type) for the dataset."""
        stems = [fp.stem.replace("_embeddings", "") for fp in self.path.glob("*_embeddings.npy")]
        return {tuple(stem.split("_")) for stem in stems}  # type: ignore

    @property
    def available_columns(self) -> List[str]:
        """Return the available columns in the dataset."""
        available_columns = pd.read_csv(self.csv_path, nrows=0).columns.tolist()

        # filter in case the column names include the index
        available_columns = list(filter(lambda x: x != "Unnamed: 0", available_columns))

        return available_columns

    def to_npy_arrays(
        self, input_names: List[str], target_names: Optional[List[str]] = None
    ) -> Tuple[List[np.ndarray], Optional[List[np.ndarray]]]:
        """Load dataset inputs and targets into numpy arrays.

        Args:
            input_names: list of input column names
            target_names: list of target column names

        Return:
            list of inputs arrays and targets arrays

        Raises:
            ValueError: one or several columns are not available
        """
        columns_to_load = input_names + target_names if target_names is not None else input_names

        available_columns = set(pd.read_csv(self.csv_path, nrows=1).columns)
        not_available_columns = set(columns_to_load) - available_columns

        if not_available_columns:
            raise ValueError(f"The following columns are not available: {not_available_columns}")

        df = pd.read_csv(self.csv_path, usecols=columns_to_load)

        inputs = [df[input_name].values for input_name in input_names]

        targets = None
        if target_names is not None:
            targets = [df[target_name].values for target_name in target_names]

        return inputs, targets

    def get_embeddings(
        self,
        variable_name: str,
        model_name: str = "esm-1b",
        embeddings_type: str = "cls",
    ) -> np.ndarray:
        """Return a 2D numpy array with the pretrained embeddings for each sequence.

        Args:
            variable_name: name of the sequence variable
            model_name: name of the model from which come the embeddings
            embeddings_type: type of the embeddings

        Return:
            list of inputs arrays and targets arrays

        Raises:
            ValueError: the embeddings are not available
        """
        if (
            variable_name,
            model_name,
            embeddings_type,
        ) not in self.available_embeddings:
            msg = f"""The embeddings for the sequence {variable_name} with model {model_name} and type {embeddings_type}
            are not available.
            Embeddings available: {self.available_embeddings}
            """
            raise ValueError(msg)

        embeddings = np.load(
            self.path / f"{variable_name}_{model_name}_{embeddings_type}_embeddings.npy"
        )

        return embeddings

    def display_description(self) -> None:
        """Display the description of the dataset."""
        with open(self.description_path, "r") as description_file:
            print(description_file.read())


def list_datasets(include_tests: bool = False) -> List[str]:
    """List all the datasets in the bucket.

    Args:
        include_tests: include test datasets in the list

    Return:
        list of the datasets
    """
    blobs = list_blobs()
    dataset_names = list(set(map(lambda x: x.name.split("/")[0], blobs)))
    if not include_tests:
        dataset_names = list(filter(lambda x: not x.startswith("test"), dataset_names))

    return dataset_names


def load_dataset(name: str, force: bool = False) -> Optional[Dataset]:
    """Load a bio-dataset.

    Args:
        name: name of the dataset
        force: force fetching the dataset

    Return:
        a Dataset instance
    """
    available_datasets = list_datasets(include_tests="test" in name)

    if name not in available_datasets:
        log.error(f"Dataset {name} does not exist.")
        return None

    return Dataset(name, force=force)



"""Define all the useful directories for the project.

- ROOT_DIRECTORY: main directory

- CURRENT_DIRECTORY: directory with all the code

- HOME_DIRECTORY: home directory

- CACHE_DIRECTORY: directory used to store the cache datasets
"""
from pathlib import Path

CURRENT_DIRECTORY = Path(__file__).resolve().parent
ROOT_DIRECTORY = CURRENT_DIRECTORY.parent

HOME_DIRECTORY = Path.home()
CACHE_DIRECTORY = HOME_DIRECTORY / ".cache" / "bio-datasets"

from .dataset import list_datasets  # noqa: E402
from .dataset import load_dataset  # noqa: E402



from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import patch
from unittest.mock import PropertyMock

import pytest
from google.cloud.storage import Blob

from biodatasets import list_datasets
from biodatasets.dataset import Dataset
from biodatasets.utils.google_bucket import DATASET_BUCKET_NAME


@pytest.fixture()
def dataset_description_path(tmp_path_factory) -> Path:
    """Fixture to define the dataset description_path."""
    dataset_path = tmp_path_factory.mktemp("test_dataset") / "description.md"
    description = "This is a fake description of the dataset.\nAnother line."
    with open(dataset_path, "w") as file:
        file.write(description)

    return dataset_path


@patch("biodatasets.dataset.pull_dataset")
@patch("biodatasets.dataset.Dataset.description_path", new_callable=PropertyMock)
def test_display_description_describe(
    desc_path_mock: MagicMock, pull_dataset_mock: MagicMock, dataset_description_path, capfd
):
    # mock pull_dataset to avoid loading a dataset
    pull_dataset_mock.return_value = None

    desc_path_mock.return_value = dataset_description_path
    dataset = Dataset("test")
    dataset.display_description()

    out, _ = capfd.readouterr()
    assert out == "This is a fake description of the dataset.\nAnother line.\n"


@patch("biodatasets.dataset.list_blobs")
@pytest.mark.parametrize("include_tests", [False, True])
def test_list_datasets(list_blobs_mock: MagicMock, include_tests: bool):
    list_blobs_mock.return_value = [
        Blob(name=x, bucket=DATASET_BUCKET_NAME)
        for x in [
            "test_1/dataset.csv",
            "pathogen",
            "pathogen/description.md",
            "protein-binding/info.json",
            "test2",
        ]
    ]

    datasets = list_datasets(include_tests)

    list_blobs_mock.assert_called_once()
    if include_tests:
        assert set(datasets) == {"test_1", "pathogen", "protein-binding", "test2"}
    else:
        assert set(datasets) == {"pathogen", "protein-binding"}



"""Package setup file"""
from typing import List

from setuptools import find_packages
from setuptools import setup

from biodatasets import ROOT_DIRECTORY
from biodatasets.version import VERSION

README = (ROOT_DIRECTORY / "README.md").read_text()


def read_requirements() -> List:
    with open("requirements.txt", "r+") as file:
        req = file.readlines()
    requirements = [pkg.replace("\n", "") for pkg in req]

    return requirements


DESCRIPTION = "Open-source collection of biology datasets and pre-trained embeddings."


def make_install():
    """main install function"""

    setup_fn = setup(
        name="bio-datasets",
        license="Apache-2.0",
        version=VERSION,
        description=DESCRIPTION,
        author="InstaDeep",
        long_description=README,
        long_description_content_type="text/markdown",
        url="https://github.com/DeepChainBio/datasets",
        author_email="a.delfosse@instadeep.com",
        packages=find_packages(exclude=["test"]),
        classifiers=[
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3.7",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "Topic :: Software Development",
        ],
        install_requires=read_requirements(),
        include_package_data=True,
        zip_safe=False,
        python_requires=">=3.7",
    )

    return setup_fn


if __name__ == "__main__":
    make_install()
